@@ -34,9 +34,13 @@
 #include <errno.h>
 #include <sys/stat.h>
 #include <sys/ioctl.h>
-#include "drm.h"
-#include "i915_drm.h"
+
+#include <drm.h>
+
+#include "ioctl_wrappers.h"
 #include "drmtest.h"
+#include "igt_debugfs.h"
+#include "igt_aux.h"
 
 #define OBJECT_SIZE (16*1024*1024)
 
@@ -78,11 +82,11 @@ test_fence_restore(int fd, bool tiled2untiled)
 
 	igt_system_suspend_autoresume();
 
-	printf("checking the first canary object\n");
+	igt_info("checking the first canary object\n");
 	for (i = 0; i < OBJECT_SIZE/sizeof(uint32_t); i++)
 		igt_assert(ptr1[i] == i);
 
-	printf("checking the second canary object\n");
+	igt_info("checking the second canary object\n");
 	for (i = 0; i < OBJECT_SIZE/sizeof(uint32_t); i++)
 		igt_assert(ptr2[i] == i);
 
@@ -108,7 +112,7 @@ test_debugfs_reader(void)
 		snprintf(tmp, sizeof(tmp) - 1,
 			 "while true; do find %s/%i/ -type f | xargs cat > /dev/null 2>&1; done",
 			 dfs_base, drm_get_card());
-		assert(execl("/bin/sh", "sh", "-c", tmp, (char *) NULL) != -1);
+		igt_assert(execl("/bin/sh", "sh", "-c", tmp, (char *) NULL) != -1);
 	}
 
 	sleep(1);
@@ -133,7 +137,7 @@ test_sysfs_reader(void)
 		snprintf(tmp, sizeof(tmp) - 1,
 			 "while true; do find %s%i*/ -type f | xargs cat > /dev/null 2>&1; done",
 			 dfs_base, drm_get_card());
-		assert(execl("/bin/sh", "sh", "-c", tmp, (char *) NULL) != -1);
+		igt_assert(execl("/bin/sh", "sh", "-c", tmp, (char *) NULL) != -1);
 	}
 
 	sleep(1);
@@ -145,6 +149,17 @@ test_sysfs_reader(void)
 	igt_stop_helper(&reader);
 }
 
+static void
+test_forcewake(void)
+{
+	int fw_fd;
+
+	fw_fd = igt_open_forcewake_handle();
+	igt_assert(fw_fd >= 0);
+	igt_system_suspend_autoresume();
+	close (fw_fd);
+}
+
 int fd;
 
 igt_main
@@ -166,6 +181,9 @@ igt_main
 	igt_subtest("sysfs-reader")
 		test_sysfs_reader();
 
+	igt_subtest("forcewake")
+		test_forcewake();
+
 	igt_fixture
 		close(fd);
 }
