@@ -51,6 +51,9 @@
 #include "igt_kms.h"
 #include "igt_debugfs.h"
 
+/* One day, this will be on your libdrm. */
+#define DRM_CLIENT_CAP_UNIVERSAL_PLANES 2
+
 #define MSR_PC8_RES	0x630
 #define MSR_PC9_RES	0x631
 #define MSR_PC10_RES	0x632
@@ -72,6 +75,12 @@ enum screen_type {
 	SCREEN_TYPE_ANY,
 };
 
+enum plane_type {
+	PLANE_OVERLAY,
+	PLANE_PRIMARY,
+	PLANE_CURSOR,
+};
+
 /* Wait flags */
 #define DONT_WAIT	0
 #define WAIT_STATUS	1
@@ -82,7 +91,6 @@ enum screen_type {
 int drm_fd, msr_fd, pm_status_fd, pc8_status_fd;
 bool has_runtime_pm, has_pc8;
 struct mode_set_data ms_data;
-struct scanout_fb *fbs = NULL;
 
 /* Stuff used when creating FBs and mode setting. */
 struct mode_set_data {
@@ -102,15 +110,17 @@ struct compare_data {
 	drmModePropertyBlobPtr edids[MAX_CONNECTORS];
 };
 
-/* During the stress tests we want to be as fast as possible, so use pre-created
- * FBs instead of creating them again and again. */
-struct scanout_fb {
-	uint32_t handle;
-	int width;
-	int height;
-	struct scanout_fb *next;
+struct modeset_params {
+	uint32_t crtc_id;
+	uint32_t connector_id;
+	struct igt_fb fb;
+	drmModeModeInfoPtr mode;
 };
 
+struct modeset_params lpsp_mode_params;
+struct modeset_params non_lpsp_mode_params;
+struct modeset_params *default_mode_params;
+
 /* If the read fails, then the machine doesn't support PC8+ residencies. */
 static bool supports_pc8_plus_residencies(void)
 {
@@ -223,66 +233,35 @@ static void disable_all_screens_dpms(struct mode_set_data *data)
 
 static void disable_all_screens(struct mode_set_data *data)
 {
-	int i, rc;
-
-	for (i = 0; i < data->res->count_crtcs; i++) {
-		rc = drmModeSetCrtc(drm_fd, data->res->crtcs[i], -1, 0, 0,
-				    NULL, 0, NULL);
-		igt_assert(rc == 0);
-	}
+	kmstest_unset_all_crtcs(drm_fd, data->res);
 }
 
-static struct scanout_fb *create_fb(struct mode_set_data *data, int width,
-				    int height)
+#define disable_all_screens_and_wait(data) do { \
+	disable_all_screens(data); \
+	igt_assert(wait_for_suspended()); \
+} while (0)
+
+static void disable_or_dpms_all_screens(struct mode_set_data *data, bool dpms)
 {
-	struct scanout_fb *fb_info;
-	struct igt_fb fb;
-	cairo_t *cr;
-
-	fb_info = malloc(sizeof(struct scanout_fb));
-	igt_assert(fb_info);
-
-	fb_info->handle = igt_create_fb(drm_fd, width, height,
-					    DRM_FORMAT_XRGB8888,
-					    false, &fb);
-	fb_info->width = width;
-	fb_info->height = height;
-	fb_info->next = NULL;
-
-	cr = igt_get_cairo_ctx(drm_fd, &fb);
-	igt_paint_test_pattern(cr, width, height);
-	cairo_destroy(cr);
-
-	return fb_info;
+	if (dpms)
+		disable_all_screens_dpms(&ms_data);
+	else
+		disable_all_screens(&ms_data);
 }
 
-static uint32_t get_fb(struct mode_set_data *data, int width, int height)
+#define disable_or_dpms_all_screens_and_wait(data, dpms) do { \
+	disable_or_dpms_all_screens((data), (dpms)); \
+	igt_assert(wait_for_suspended()); \
+} while (0)
+
+static bool init_modeset_params_for_type(struct mode_set_data *data,
+					 struct modeset_params *params,
+					 enum screen_type type)
 {
-	struct scanout_fb *fb;
-
-	if (!fbs) {
-		fbs = create_fb(data, width, height);
-		return fbs->handle;
-	}
-
-	for (fb = fbs; fb != NULL; fb = fb->next) {
-		if (fb->width == width && fb->height == height)
-			return fb->handle;
-
-		if (!fb->next) {
-			fb->next = create_fb(data, width, height);
-			return fb->next->handle;
-		}
-	}
-	igt_assert(false);
-}
-
-static bool enable_one_screen_with_type(struct mode_set_data *data,
-					enum screen_type type)
-{
-	uint32_t crtc_id = 0, buffer_id = 0, connector_id = 0;
+	int i;
+	uint32_t connector_id = 0;
 	drmModeModeInfoPtr mode = NULL;
-	int i, rc;
+	cairo_t *cr;
 
 	for (i = 0; i < data->res->count_connectors; i++) {
 		drmModeConnectorPtr c = data->connectors[i];
@@ -302,56 +281,111 @@ static bool enable_one_screen_with_type(struct mode_set_data *data,
 		}
 	}
 
-	if (connector_id == 0)
+	if (!connector_id)
 		return false;
 
-	crtc_id = data->res->crtcs[0];
-	buffer_id = get_fb(data, mode->hdisplay, mode->vdisplay);
+	igt_create_fb(drm_fd, mode->hdisplay, mode->vdisplay,
+		      DRM_FORMAT_XRGB8888, I915_TILING_NONE, &params->fb);
+	cr = igt_get_cairo_ctx(drm_fd, &params->fb);
+	igt_paint_test_pattern(cr, mode->hdisplay, mode->vdisplay);
+	cairo_destroy(cr);
 
-	igt_assert(crtc_id);
-	igt_assert(buffer_id);
-	igt_assert(connector_id);
-	igt_assert(mode);
-
-	rc = drmModeSetCrtc(drm_fd, crtc_id, buffer_id, 0, 0, &connector_id,
-			    1, mode);
-	igt_assert(rc == 0);
+	params->crtc_id = data->res->crtcs[0];
+	params->connector_id = connector_id;
+	params->mode = mode;
 
 	return true;
 }
 
+static void init_modeset_cached_params(struct mode_set_data *data)
+{
+	bool lpsp, non_lpsp;
+
+	lpsp = init_modeset_params_for_type(data, &lpsp_mode_params,
+					    SCREEN_TYPE_LPSP);
+	non_lpsp = init_modeset_params_for_type(data, &non_lpsp_mode_params,
+						SCREEN_TYPE_NON_LPSP);
+
+	if (lpsp)
+		default_mode_params = &lpsp_mode_params;
+	else if (non_lpsp)
+		default_mode_params = &non_lpsp_mode_params;
+	else
+		default_mode_params = NULL;
+}
+
+static bool set_mode_for_params(struct modeset_params *params)
+{
+	int rc;
+
+	rc = drmModeSetCrtc(drm_fd, params->crtc_id, params->fb.fb_id, 0, 0,
+			    &params->connector_id, 1, params->mode);
+	return (rc == 0);
+}
+
+#define set_mode_for_params_and_wait(params) do { \
+	igt_assert(set_mode_for_params(params)); \
+	igt_assert(wait_for_active()); \
+} while (0)
+
+static bool enable_one_screen_with_type(struct mode_set_data *data,
+					enum screen_type type)
+{
+	struct modeset_params *params = NULL;
+
+	switch (type) {
+	case SCREEN_TYPE_ANY:
+		params = default_mode_params;
+		break;
+	case SCREEN_TYPE_LPSP:
+		params = &lpsp_mode_params;
+		break;
+	case SCREEN_TYPE_NON_LPSP:
+		params = &non_lpsp_mode_params;
+		break;
+	default:
+		igt_assert(0);
+	}
+
+	if (!params)
+		return false;
+
+	return set_mode_for_params(params);
+}
+
 static void enable_one_screen(struct mode_set_data *data)
 {
 	/* SKIP if there are no connected screens. */
 	igt_require(enable_one_screen_with_type(data, SCREEN_TYPE_ANY));
 }
 
+#define enable_one_screen_and_wait(data) do { \
+	enable_one_screen(data); \
+	igt_assert(wait_for_active()); \
+} while (0)
+
 static drmModePropertyBlobPtr get_connector_edid(drmModeConnectorPtr connector,
 						 int index)
 {
-	unsigned int i;
-	drmModeObjectPropertiesPtr props;
-	drmModePropertyBlobPtr ret = NULL;
+	bool found;
+	uint64_t prop_value;
+	drmModePropertyPtr prop;
+	drmModePropertyBlobPtr blob = NULL;
 
-	props = drmModeObjectGetProperties(drm_fd, connector->connector_id,
-					   DRM_MODE_OBJECT_CONNECTOR);
+	found = kmstest_get_property(drm_fd, connector->connector_id,
+				     DRM_MODE_OBJECT_CONNECTOR, "EDID",
+				     NULL, &prop_value, &prop);
 
-	for (i = 0; i < props->count_props; i++) {
-		drmModePropertyPtr prop = drmModeGetProperty(drm_fd,
-							     props->props[i]);
+	if (found) {
+		igt_assert(prop->flags & DRM_MODE_PROP_BLOB);
+		igt_assert(prop->count_blobs == 0);
 
-		if (strcmp(prop->name, "EDID") == 0) {
-			igt_assert(prop->flags & DRM_MODE_PROP_BLOB);
-			igt_assert(prop->count_blobs == 0);
-			ret = drmModeGetPropertyBlob(drm_fd,
-						     props->prop_values[i]);
-		}
+		blob = drmModeGetPropertyBlob(drm_fd, prop_value);
 
 		drmModeFreeProperty(prop);
 	}
 
-	drmModeFreeObjectProperties(props);
-	return ret;
+	return blob;
 }
 
 static void init_mode_set_data(struct mode_set_data *data)
@@ -370,7 +404,9 @@ static void init_mode_set_data(struct mode_set_data *data)
 
 	data->devid = intel_get_drm_devid(drm_fd);
 
-	igt_set_vt_graphics_mode();
+	kmstest_set_vt_graphics_mode();
+
+	init_modeset_cached_params(&ms_data);
 }
 
 static void fini_mode_set_data(struct mode_set_data *data)
@@ -605,7 +641,7 @@ static void test_i2c(struct mode_set_data *data)
 	int i2c_edids = count_i2c_valid_edids();
 	int drm_edids = count_drm_valid_edids(data);
 
-	igt_assert_cmpint(i2c_edids, ==, drm_edids);
+	igt_assert_eq(i2c_edids, drm_edids);
 }
 
 static void setup_pc8(void)
@@ -694,19 +730,11 @@ static void setup_environment(void)
 	igt_info("PC8 residency support: %d\n", has_pc8);
 
 	igt_require(has_runtime_pm);
+
 }
 
 static void teardown_environment(void)
 {
-	struct scanout_fb *fb, *fb_next;
-
-	fb = fbs;
-	while (fb) {
-		fb_next = fb->next;
-		free(fb);
-		fb = fb_next;
-	}
-
 	fini_mode_set_data(&ms_data);
 	drmClose(drm_fd);
 	close(msr_fd);
@@ -716,11 +744,9 @@ static void teardown_environment(void)
 
 static void basic_subtest(void)
 {
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
-	enable_one_screen(&ms_data);
-	igt_assert(wait_for_active());
+	enable_one_screen_and_wait(&ms_data);
 }
 
 static void pc8_residency_subtest(void)
@@ -779,18 +805,15 @@ static void drm_resources_equal_subtest(void)
 {
 	struct compare_data pre_suspend, during_suspend, post_suspend;
 
-	enable_one_screen(&ms_data);
-	igt_assert(wait_for_active());
+	enable_one_screen_and_wait(&ms_data);
 	get_drm_info(&pre_suspend);
 	igt_assert(wait_for_active());
 
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 	get_drm_info(&during_suspend);
 	igt_assert(wait_for_suspended());
 
-	enable_one_screen(&ms_data);
-	igt_assert(wait_for_active());
+	enable_one_screen_and_wait(&ms_data);
 	get_drm_info(&post_suspend);
 	igt_assert(wait_for_active());
 
@@ -826,11 +849,9 @@ static void i2c_subtest(void)
 {
 	i2c_subtest_check_environment();
 
-	enable_one_screen(&ms_data);
-	igt_assert(wait_for_active());
+	enable_one_screen_and_wait(&ms_data);
 
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 	test_i2c(&ms_data);
 	igt_assert(wait_for_suspended());
 
@@ -908,8 +929,7 @@ static void debugfs_read_subtest(void)
 	igt_require_f(dir, "Can't open the debugfs directory\n");
 	closedir(dir);
 
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	read_files_from_dir(path, 0);
 }
@@ -924,8 +944,7 @@ static void sysfs_read_subtest(void)
 	igt_require_f(dir, "Can't open the sysfs directory\n");
 	closedir(dir);
 
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	read_files_from_dir(path, 0);
 }
@@ -937,8 +956,7 @@ static void debugfs_forcewake_user_subtest(void)
 
 	igt_require(intel_gen(ms_data.devid) >= 6);
 
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	fd = igt_open_forcewake_handle();
 	igt_require(fd >= 0);
@@ -965,8 +983,7 @@ static void gem_mmap_subtest(bool gtt_mmap)
 	uint8_t *gem_buf;
 
 	/* Create, map and set data while the device is active. */
-	enable_one_screen(&ms_data);
-	igt_assert(wait_for_active());
+	enable_one_screen_and_wait(&ms_data);
 
 	handle = gem_create(drm_fd, buf_size);
 
@@ -984,8 +1001,7 @@ static void gem_mmap_subtest(bool gtt_mmap)
 		igt_assert(gem_buf[i] == (i & 0xFF));
 
 	/* Now suspend, read and modify. */
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	for (i = 0; i < buf_size; i++)
 		igt_assert(gem_buf[i] == (i & 0xFF));
@@ -996,8 +1012,7 @@ static void gem_mmap_subtest(bool gtt_mmap)
 	igt_assert(wait_for_suspended());
 
 	/* Now resume and see if it's still there. */
-	enable_one_screen(&ms_data);
-	igt_assert(wait_for_active());
+	enable_one_screen_and_wait(&ms_data);
 	for (i = 0; i < buf_size; i++)
 		igt_assert(gem_buf[i] == (~i & 0xFF));
 
@@ -1005,8 +1020,7 @@ static void gem_mmap_subtest(bool gtt_mmap)
 
 	/* Now the opposite: suspend, and try to create the mmap while
 	 * suspended. */
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	if (gtt_mmap)
 		gem_buf = gem_mmap__gtt(drm_fd, handle, buf_size,
@@ -1025,8 +1039,7 @@ static void gem_mmap_subtest(bool gtt_mmap)
 	igt_assert(wait_for_suspended());
 
 	/* Resume and check if it's still there. */
-	enable_one_screen(&ms_data);
-	igt_assert(wait_for_active());
+	enable_one_screen_and_wait(&ms_data);
 	for (i = 0; i < buf_size; i++)
 		igt_assert(gem_buf[i] == (i & 0xFF));
 
@@ -1049,8 +1062,7 @@ static void gem_pread_subtest(void)
 	memset(read_buf, 0, buf_size);
 
 	/* Create and set data while the device is active. */
-	enable_one_screen(&ms_data);
-	igt_assert(wait_for_active());
+	enable_one_screen_and_wait(&ms_data);
 
 	handle = gem_create(drm_fd, buf_size);
 
@@ -1065,8 +1077,7 @@ static void gem_pread_subtest(void)
 		igt_assert(cpu_buf[i] == read_buf[i]);
 
 	/* Now suspend, read and modify. */
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	memset(read_buf, 0, buf_size);
 	gem_read(drm_fd, handle, 0, read_buf, buf_size);
@@ -1081,8 +1092,7 @@ static void gem_pread_subtest(void)
 	igt_assert(wait_for_suspended());
 
 	/* Now resume and see if it's still there. */
-	enable_one_screen(&ms_data);
-	igt_assert(wait_for_active());
+	enable_one_screen_and_wait(&ms_data);
 
 	memset(read_buf, 0, buf_size);
 	gem_read(drm_fd, handle, 0, read_buf, buf_size);
@@ -1189,8 +1199,7 @@ static void gem_execbuf_subtest(void)
 	uint32_t color;
 
 	/* Create and set data while the device is active. */
-	enable_one_screen(&ms_data);
-	igt_assert(wait_for_active());
+	enable_one_screen_and_wait(&ms_data);
 
 	handle = gem_create(drm_fd, dst_size);
 
@@ -1200,8 +1209,7 @@ static void gem_execbuf_subtest(void)
 	gem_write(drm_fd, handle, 0, cpu_buf, dst_size);
 
 	/* Now suspend and try it. */
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	color = 0x12345678;
 	submit_blt_cmd(handle, sq_x, sq_y, sq_w, sq_h, pitch, color,
@@ -1223,8 +1231,7 @@ static void gem_execbuf_subtest(void)
 	}
 
 	/* Now resume and check for it again. */
-	enable_one_screen(&ms_data);
-	igt_assert(wait_for_active());
+	enable_one_screen_and_wait(&ms_data);
 
 	memset(cpu_buf, 0, dst_size);
 	gem_read(drm_fd, handle, 0, cpu_buf, dst_size);
@@ -1247,8 +1254,7 @@ static void gem_execbuf_subtest(void)
 	submit_blt_cmd(handle, sq_x, sq_y, sq_w, sq_h, pitch, color,
 		       &presumed_offset);
 
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	memset(cpu_buf, 0, dst_size);
 	gem_read(drm_fd, handle, 0, cpu_buf, dst_size);
@@ -1290,8 +1296,7 @@ static void gem_execbuf_stress_subtest(int rounds, int wait_flags)
 	batch_buf[i++] = MI_NOOP;
 	igt_assert(i * sizeof(uint32_t) == batch_size);
 
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	handle = gem_create(drm_fd, batch_size);
 	gem_write(drm_fd, handle, 0, batch_buf, batch_size);
@@ -1321,8 +1326,7 @@ static void gem_execbuf_stress_subtest(int rounds, int wait_flags)
 /* When this test was written, it triggered WARNs and DRM_ERRORs on dmesg. */
 static void gem_idle_subtest(void)
 {
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	sleep(5);
 
@@ -1336,8 +1340,7 @@ static void reg_read_ioctl_subtest(void)
 		.offset = 0x2358, /* render ring timestamp */
 	};
 
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	do_ioctl(drm_fd, DRM_IOCTL_I915_REG_READ, &rr);
 
@@ -1362,21 +1365,18 @@ static void pci_d3_state_subtest(void)
 {
 	igt_require(has_runtime_pm);
 
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	igt_assert(device_in_pci_d3());
 
-	enable_one_screen(&ms_data);
-	igt_assert(wait_for_active());
+	enable_one_screen_and_wait(&ms_data);
 
 	igt_assert(!device_in_pci_d3());
 }
 
 static void stay_subtest(void)
 {
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 
 	while (1)
 		sleep(600);
@@ -1384,30 +1384,377 @@ static void stay_subtest(void)
 
 static void system_suspend_subtest(void)
 {
-	disable_all_screens(&ms_data);
-	igt_assert(wait_for_suspended());
+	disable_all_screens_and_wait(&ms_data);
 	igt_system_suspend_autoresume();
 	igt_assert(wait_for_suspended());
 }
 
+/* Enable a screen, activate DPMS, then do a modeset. At some point our driver
+ * produced WARNs on this case. */
+static void dpms_mode_unset_subtest(enum screen_type type)
+{
+	disable_all_screens_and_wait(&ms_data);
+
+	igt_require(enable_one_screen_with_type(&ms_data, type));
+	igt_assert(wait_for_active());
+
+	disable_all_screens_dpms(&ms_data);
+	igt_assert(wait_for_suspended());
+
+	disable_all_screens_and_wait(&ms_data);
+}
+
+static void fill_igt_fb(struct igt_fb *fb, uint32_t color)
+{
+	int i;
+	uint32_t *ptr;
+
+	ptr = gem_mmap__gtt(drm_fd, fb->gem_handle, fb->size, PROT_WRITE);
+	for (i = 0; i < fb->size/sizeof(uint32_t); i++)
+		ptr[i] = color;
+	igt_assert(munmap(ptr, fb->size) == 0);
+}
+
+/* At some point, this test triggered WARNs in the Kernel. */
+static void cursor_subtest(bool dpms)
+{
+	int rc;
+	struct igt_fb cursor_fb1, cursor_fb2, cursor_fb3;
+	uint32_t crtc_id;
+
+	disable_all_screens_and_wait(&ms_data);
+
+	igt_require(default_mode_params);
+	crtc_id = default_mode_params->crtc_id;
+
+	igt_create_fb(drm_fd, 64, 64, DRM_FORMAT_ARGB8888, I915_TILING_NONE,
+		      &cursor_fb1);
+	igt_create_fb(drm_fd, 64, 64, DRM_FORMAT_ARGB8888, I915_TILING_NONE,
+		      &cursor_fb2);
+	igt_create_fb(drm_fd, 64, 64, DRM_FORMAT_ARGB8888, I915_TILING_X,
+		      &cursor_fb3);
+
+	fill_igt_fb(&cursor_fb1, 0xFF00FFFF);
+	fill_igt_fb(&cursor_fb2, 0xFF00FF00);
+	fill_igt_fb(&cursor_fb3, 0xFFFF0000);
+
+	set_mode_for_params_and_wait(default_mode_params);
+
+	rc = drmModeSetCursor(drm_fd, crtc_id, cursor_fb1.gem_handle,
+			      cursor_fb1.width, cursor_fb1.height);
+	igt_assert(rc == 0);
+	rc = drmModeMoveCursor(drm_fd, crtc_id, 0, 0);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_active());
+
+	disable_or_dpms_all_screens_and_wait(&ms_data, dpms);
+
+	/* First, just move the cursor. */
+	rc = drmModeMoveCursor(drm_fd, crtc_id, 1, 1);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_suspended());
+
+	/* Then unset it, and set a new one. */
+	rc = drmModeSetCursor(drm_fd, crtc_id, 0, 0, 0);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_suspended());
+
+	rc = drmModeSetCursor(drm_fd, crtc_id, cursor_fb2.gem_handle,
+			      cursor_fb1.width, cursor_fb2.height);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_suspended());
+
+	/* Move the new cursor. */
+	rc = drmModeMoveCursor(drm_fd, crtc_id, 2, 2);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_suspended());
+
+	/* Now set a new one without unsetting the previous one. */
+	rc = drmModeSetCursor(drm_fd, crtc_id, cursor_fb1.gem_handle,
+			      cursor_fb1.width, cursor_fb1.height);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_suspended());
+
+	/* Cursor 3 was created with tiling and painted with a GTT mmap, so
+	 * hopefully it has some fences around it. */
+	rc = drmModeRmFB(drm_fd, cursor_fb3.fb_id);
+	igt_assert(rc == 0);
+	gem_set_tiling(drm_fd, cursor_fb3.gem_handle, false, cursor_fb3.stride);
+	igt_assert(wait_for_suspended());
+
+	rc = drmModeSetCursor(drm_fd, crtc_id, cursor_fb3.gem_handle,
+			      cursor_fb3.width, cursor_fb3.height);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_suspended());
+
+	/* Make sure nothing remains for the other tests. */
+	rc = drmModeSetCursor(drm_fd, crtc_id, 0, 0, 0);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_suspended());
+}
+
+static enum plane_type get_plane_type(uint32_t plane_id)
+{
+	int i;
+	bool found;
+	uint64_t prop_value;
+	drmModePropertyPtr prop;
+	const char *enum_name = NULL;
+	enum plane_type type;
+
+	found = kmstest_get_property(drm_fd, plane_id, DRM_MODE_OBJECT_PLANE,
+				     "type", NULL, &prop_value, &prop);
+	igt_assert(found);
+
+	igt_assert(prop->flags & DRM_MODE_PROP_ENUM);
+	igt_assert(prop_value < prop->count_enums);
+
+	for (i = 0; i < prop->count_enums; i++) {
+		if (prop->enums[i].value == prop_value) {
+			enum_name = prop->enums[i].name;
+			break;
+		}
+	}
+	igt_assert(enum_name);
+
+	if (strcmp(enum_name, "Overlay") == 0)
+		type = PLANE_OVERLAY;
+	else if (strcmp(enum_name, "Primary") == 0)
+		type = PLANE_PRIMARY;
+	else if (strcmp(enum_name, "Cursor") == 0)
+		type = PLANE_CURSOR;
+	else
+		igt_assert(0);
+
+	drmModeFreeProperty(prop);
+
+	return type;
+}
+
+static void test_one_plane(bool dpms, uint32_t plane_id,
+			   enum plane_type plane_type)
+{
+	int rc;
+	uint32_t plane_format, plane_w, plane_h;
+	uint32_t crtc_id;
+	struct igt_fb plane_fb1, plane_fb2;
+	int32_t crtc_x = 0, crtc_y = 0;
+	unsigned int tiling;
+
+	disable_all_screens_and_wait(&ms_data);
+
+	igt_require(default_mode_params);
+	crtc_id = default_mode_params->crtc_id;
+
+	switch (plane_type) {
+	case PLANE_OVERLAY:
+		plane_format = DRM_FORMAT_XRGB8888;
+		plane_w = 64;
+		plane_h = 64;
+		tiling = I915_TILING_X;
+		break;
+	case PLANE_PRIMARY:
+		plane_format = DRM_FORMAT_XRGB8888;
+		plane_w = default_mode_params->mode->hdisplay;
+		plane_h = default_mode_params->mode->vdisplay;
+		tiling = I915_TILING_X;
+		break;
+	case PLANE_CURSOR:
+		plane_format = DRM_FORMAT_ARGB8888;
+		plane_w = 64;
+		plane_h = 64;
+		tiling = I915_TILING_NONE;
+		break;
+	default:
+		igt_assert(0);
+		break;
+	}
+
+	igt_create_fb(drm_fd, plane_w, plane_h, plane_format, tiling,
+		      &plane_fb1);
+	igt_create_fb(drm_fd, plane_w, plane_h, plane_format, tiling,
+		      &plane_fb2);
+	fill_igt_fb(&plane_fb1, 0xFF00FFFF);
+	fill_igt_fb(&plane_fb2, 0xFF00FF00);
+
+	set_mode_for_params_and_wait(default_mode_params);
+
+	rc = drmModeSetPlane(drm_fd, plane_id, crtc_id, plane_fb1.fb_id, 0,
+			     0, 0, plane_fb1.width, plane_fb1.height,
+			     0 << 16, 0 << 16, plane_fb1.width << 16,
+			     plane_fb1.height << 16);
+	igt_assert(rc == 0);
+
+	disable_or_dpms_all_screens_and_wait(&ms_data, dpms);
+
+	/* Just move the plane around. */
+	if (plane_type != PLANE_PRIMARY) {
+		crtc_x++;
+		crtc_y++;
+	}
+	rc = drmModeSetPlane(drm_fd, plane_id, crtc_id, plane_fb1.fb_id, 0,
+			     crtc_x, crtc_y, plane_fb1.width, plane_fb1.height,
+			     0 << 16, 0 << 16, plane_fb1.width << 16,
+			     plane_fb1.height << 16);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_suspended());
+
+	/* Unset, then change the plane. */
+	rc = drmModeSetPlane(drm_fd, plane_id, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_suspended());
+
+	rc = drmModeSetPlane(drm_fd, plane_id, crtc_id, plane_fb2.fb_id, 0,
+			     crtc_x, crtc_y, plane_fb2.width, plane_fb2.height,
+			     0 << 16, 0 << 16, plane_fb2.width << 16,
+			     plane_fb2.height << 16);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_suspended());
+
+	/* Now change the plane without unsetting first. */
+	rc = drmModeSetPlane(drm_fd, plane_id, crtc_id, plane_fb1.fb_id, 0,
+			     crtc_x, crtc_y, plane_fb1.width, plane_fb1.height,
+			     0 << 16, 0 << 16, plane_fb1.width << 16,
+			     plane_fb1.height << 16);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_suspended());
+
+	/* Make sure nothing remains for the other tests. */
+	rc = drmModeSetPlane(drm_fd, plane_id, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);
+	igt_assert(rc == 0);
+	igt_assert(wait_for_suspended());
+}
+
+/* This one also triggered WARNs on our driver at some point in time. */
+static void planes_subtest(bool universal, bool dpms)
+{
+	int i, rc, planes_tested = 0;
+	drmModePlaneResPtr planes;
+
+	if (universal) {
+		rc = drmSetClientCap(drm_fd, DRM_CLIENT_CAP_UNIVERSAL_PLANES,
+				     1);
+		igt_require(rc == 0);
+	}
+
+	planes = drmModeGetPlaneResources(drm_fd);
+	for (i = 0; i < planes->count_planes; i++) {
+		drmModePlanePtr plane;
+
+		plane = drmModeGetPlane(drm_fd, planes->planes[i]);
+		igt_assert(plane);
+
+		/* We just pick the first CRTC on the list, so we can test for
+		 * 0x1 as the index. */
+		if (plane->possible_crtcs & 0x1) {
+			enum plane_type type;
+
+			type = universal ? get_plane_type(plane->plane_id) :
+					   PLANE_OVERLAY;
+			test_one_plane(dpms, plane->plane_id, type);
+			planes_tested++;
+		}
+		drmModeFreePlane(plane);
+	}
+	drmModeFreePlaneResources(planes);
+
+	if (universal) {
+		rc = drmSetClientCap(drm_fd, DRM_CLIENT_CAP_UNIVERSAL_PLANES, 0);
+		igt_assert(rc == 0);
+
+		igt_assert(planes_tested >= 3);
+	} else {
+		igt_assert(planes_tested >= 1);
+	}
+}
+
+static void fences_subtest(bool dpms)
+{
+	int i;
+	uint32_t *buf_ptr;
+	uint32_t tiling = false, swizzle;
+	struct modeset_params params;
+
+	disable_all_screens_and_wait(&ms_data);
+
+	igt_require(default_mode_params);
+	params.crtc_id = default_mode_params->crtc_id;
+	params.connector_id = default_mode_params->connector_id;
+	params.mode = default_mode_params->mode;
+	igt_create_fb(drm_fd, params.mode->hdisplay, params.mode->vdisplay,
+		      DRM_FORMAT_XRGB8888, I915_TILING_X, &params.fb);
+
+	/* Even though we passed "true" as the tiling argument, double-check
+	 * that the fb is really tiled. */
+	gem_get_tiling(drm_fd, params.fb.gem_handle, &tiling, &swizzle);
+	igt_assert(tiling);
+
+	buf_ptr = gem_mmap__gtt(drm_fd, params.fb.gem_handle,
+				params.fb.size, PROT_WRITE | PROT_READ);
+	for (i = 0; i < params.fb.size/sizeof(uint32_t); i++)
+		buf_ptr[i] = i;
+
+	set_mode_for_params_and_wait(&params);
+
+	disable_or_dpms_all_screens_and_wait(&ms_data, dpms);
+
+	for (i = 0; i < params.fb.size/sizeof(uint32_t); i++)
+		igt_assert_eq(buf_ptr[i], i);
+	igt_assert(wait_for_suspended());
+
+	if (dpms) {
+		drmModeConnectorPtr c = NULL;
+
+		for (i = 0; i < ms_data.res->count_connectors; i++)
+			if (ms_data.connectors[i]->connector_id ==
+			    params.connector_id)
+				c = ms_data.connectors[i];
+		igt_assert(c);
+
+		kmstest_set_connector_dpms(drm_fd, c, DRM_MODE_DPMS_ON);
+	} else {
+		set_mode_for_params(&params);
+	}
+	igt_assert(wait_for_active());
+
+	for (i = 0; i < params.fb.size/sizeof(uint32_t); i++)
+		igt_assert_eq(buf_ptr[i], i);
+
+	igt_assert(munmap(buf_ptr, params.fb.size) == 0);
+}
+
+int rounds = 50;
+bool stay = false;
+
+static int opt_handler(int opt, int opt_index)
+{
+	switch (opt) {
+	case 'q':
+		rounds = 10;
+		break;
+	case 's':
+		stay = true;
+		break;
+	default:
+		igt_assert(0);
+	}
+
+	return 0;
+}
+
 int main(int argc, char *argv[])
 {
-	int rounds = 50;
-	bool stay = false;
+	const char *help_str =
+	       "  --quick\t\tMake the stress-tests not stressful, for quick regression testing.\n"
+	       "  --stay\t\tDisable all screen and try to go into runtime pm. Useful for debugging.";
+	static struct option long_options[] = {
+		{"quick", 0, 0, 'q'},
+		{"stay", 0, 0, 's'},
+		{ 0, 0, 0, 0 }
+	};
 
-	igt_subtest_init(argc, argv);
-
-	/* The --quick option makes the stress tests not so stressful. Useful
-	 * when you're developing and just want to make a quick test to make
-	 * sure you didn't break everything. */
-	if (argc > 1 && strcmp(argv[1], "--quick") == 0)
-		rounds = 10;
-
-	/* The --stay option enables a mode where we disable all the screens,
-	 * then stay like that, runtime suspended. This mode is useful for
-	 * running manual tests while debugging. */
-	if (argc > 1 && strcmp(argv[1], "--stay") == 0)
-		stay = true;
+	igt_subtest_init_parse_opts(argc, argv, "", long_options,
+				    help_str, opt_handler);
 
 	/* Skip instead of failing in case the machine is not prepared to reach
 	 * PC8+. We don't want bug reports from cases where the machine is just
@@ -1449,6 +1796,20 @@ int main(int argc, char *argv[])
 	igt_subtest("gem-idle")
 		gem_idle_subtest();
 
+	/* Planes and cursors */
+	igt_subtest("cursor")
+		cursor_subtest(false);
+	igt_subtest("cursor-dpms")
+		cursor_subtest(true);
+	igt_subtest("legacy-planes")
+		planes_subtest(false, false);
+	igt_subtest("legacy-planes-dpms")
+		planes_subtest(false, true);
+	igt_subtest("universal-planes")
+		planes_subtest(true, false);
+	igt_subtest("universal-planes-dpms")
+		planes_subtest(true, true);
+
 	/* Misc */
 	igt_subtest("reg-read-ioctl")
 		reg_read_ioctl_subtest();
@@ -1462,6 +1823,14 @@ int main(int argc, char *argv[])
 		debugfs_forcewake_user_subtest();
 	igt_subtest("sysfs-read")
 		sysfs_read_subtest();
+	igt_subtest("dpms-mode-unset-lpsp")
+		dpms_mode_unset_subtest(SCREEN_TYPE_LPSP);
+	igt_subtest("dpms-mode-unset-non-lpsp")
+		dpms_mode_unset_subtest(SCREEN_TYPE_NON_LPSP);
+	igt_subtest("fences")
+		fences_subtest(false);
+	igt_subtest("fences-dpms")
+		fences_subtest(true);
 
 	/* Modeset stress */
 	igt_subtest("modeset-lpsp-stress")
